from pydantic import DirectoryPath, Field

from ._basemodel import BaseModel


class WorkDirectoryModel(BaseModel):
    """The workspace defines the the root directory where all simulations are
    run. This is necessary, because some programs work with relative
    directories from some root directory.

    In this way, duqtools can ensure that the current work directory is
    a subdirectory of the given root directory. All subdirectories are
    calculated as relative to the root directory.

    For example, for `rjettov`, the root directory must be set to
    `/pfs/work/$USER/jetto/runs/`. Any UQ runs must therefore be
    a subdirectory.
    """

    root: DirectoryPath = Field(
        description='The directory from which experiments have to be run.')



from typing import List, Optional

from pydantic import Field

from ._basemodel import BaseModel
from ._description_helpers import formatter as f
from ._jetto import JettoVar


class JettoVariableModel(BaseModel):
    """Variable for describing variables specific to Jetto, The lookup table
    can be defined as a JettoVar under the lookup key."""

    type: str = Field('jetto-variable',
                      description='Discriminator for the variable type.')

    name: str = Field(description=f("""
        Name of the variable.
        Used for the lookup table to find actual fields.
        """))
    lookup: Optional[JettoVar] = Field(description=f("""
    Description of the fields that have to be updated for a Jetto Variable
    """))


class IDSVariableModel(BaseModel):
    """Variable for describing data within a IMAS database.

    The variable can be given a name, which will be used in the rest of the config
    to reference the variable. It will also be used as the column labels or
    on plots.

    The dimensions for each variable must be specified. This ensures the the data
    will be self-consistent. For example for 1D data, you can use `[x]` and for 2D data,
    `[x, y]`.

    The IDS path may contain indices. You can point to a single index, by simply giving the
    complete path (i.e. `profiles_1d/0/t_i_ave` for the 0th time slice).
    To retrieve all time slices, you can use `profiles_1d/*/t_i_ave`.
    """
    type: str = Field('IDS-variable',
                      description='discriminator for the variable type')

    name: str = Field(description=f("""
        Name of the variable.
        This will be used to reference this variable.
    """))
    ids: str = Field(description='Root IDS name.')
    path: str = Field(description=f("""
        Path to the data within the IDS.
        The fields are separated by forward slashes (`\\`).
    """))
    dims: List[str] = Field(description=f("""
        Give the dimensions of the data,
        i.e. [x] for 1D, or [x, y] for 2D data.
    """))



from __future__ import annotations

import logging
from pathlib import Path

import click
import xarray as xr

from ._plot_utils import alt_line_chart
from .config import var_lookup
from .ids import ImasHandle
from .utils import read_imas_handles_from_file

logger = logging.getLogger(__name__)
info, debug = logger.info, logger.debug


def plot(*, time_var, grid_var, data_vars, imas_paths, user, db, shot, runs,
         input_files, extensions, **kwargs):

    handle_lst = []

    for n, imas_path in enumerate(imas_paths):
        handle = ImasHandle.from_string(imas_path)
        handle_lst.append(handle)

    if db and shot and runs:
        for run in runs:
            handle_lst.append(ImasHandle(user=user, db=db, shot=shot, run=run))

    handles = {n: handle for n, handle in enumerate(handle_lst)}

    for input_file in input_files:
        handles.update(read_imas_handles_from_file(input_file))

    if len(handles) == 0:
        raise SystemExit('No data to show.')

    grid_var = var_lookup[grid_var]
    data_vars = [var_lookup[y_var] for y_var in data_vars]

    variables = [grid_var] + data_vars

    datasets = []
    for handle in handles.values():
        ds = handle.get_variables(variables=variables)
        datasets.append(ds)

    dataset = xr.concat(datasets, 'run')

    click.echo('You can now view your plot in your browser:')
    click.echo('')

    for n, y_var in enumerate(data_vars):
        click.secho(f'  {grid_var.name} vs. {y_var.name}:\n', fg='green')

        chart = alt_line_chart(dataset, x=grid_var.name, y=y_var.name)

        for extension in extensions:

            outfile = Path(f'chart_{n}.{extension}')
            click.secho(f'    file:///{outfile.absolute()}', bold=True)

            chart.save(outfile, scale_factor=2.0)

        click.echo('')



from __future__ import annotations

from ..schema.cli import ConfigModel


class Config(ConfigModel):
    """Config class containing all configs, can be used with:

        from duqtools.config import cfg
        cfg.<variable you want>

    To update the config:

        cfg.parse_file('duqtools.yaml')
    """
    _instance = None

    def __new__(cls, *args, **kwargs):
        # Make it a singleton
        if not Config._instance:
            Config._instance = object.__new__(cls)
        return Config._instance

    def parse_file(self, *args, **kwargs):
        """Add extra variables to variable lookup table."""
        super().parse_file(*args, **kwargs)
        from ._variables import var_lookup
        if self.extra_variables:
            var_lookup.update(self.extra_variables.to_variable_dict())


cfg = Config.construct()



import logging
from pathlib import Path
from typing import Any, List, Sequence

import pandas as pd

from .apply_model import apply_model
from .cleanup import remove_run
from .config import cfg
from .ids import ImasHandle
from .matrix_samplers import get_matrix_sampler
from .models import WorkDirectory
from .operations import add_to_op_queue, op_queue
from .schema.runs import Run, Runs
from .system import get_system

logger = logging.getLogger(__name__)

RUN_PREFIX = 'run_'


class CreateError(Exception):
    ...


class CreateManager:
    """Docstring for CreateManager."""

    def __init__(self):
        self.options = cfg.create

        if not self.options:
            logger.warning('No create options specified.')
            raise CreateError('No create options specified in config.')

        self.template_drc = self.options.template
        self.workspace = WorkDirectory.parse_obj(cfg.workspace)
        self.system = get_system()
        self.source = self._get_source_handle()

    def _get_source_handle(self) -> ImasHandle:
        if not self.options.template_data:
            source = self.system.imas_from_path(self.template_drc)
        else:
            source = ImasHandle.parse_obj(self.options.template_data)

        logger.info('Source data: %s', source)

        return source

    def generate_ops_list(self) -> List[Any]:
        """Generate set of operations for a run."""
        dimensions = self.options.dimensions
        matrix_sampler = get_matrix_sampler(self.options.sampler.method)
        matrix = tuple(model.expand() for model in dimensions)
        ops_list = matrix_sampler(*matrix, **dict(self.options.sampler))

        return ops_list

    def make_run_models(self, ops_list: Sequence[Any]) -> List[Run]:
        """Take list of operations and create run models."""
        run_models = []

        for i, operations in enumerate(ops_list):
            dirname = f'{RUN_PREFIX}{i:04d}'

            data_in = ImasHandle(db=self.options.data.imasdb,
                                 shot=self.source.shot,
                                 run=self.options.data.run_in_start_at + i)
            data_out = ImasHandle(db=self.options.data.imasdb,
                                  shot=self.source.shot,
                                  run=self.options.data.run_out_start_at + i)

            model = Run(dirname=dirname,
                        data_in=data_in,
                        data_out=data_out,
                        operations=operations)
            run_models.append(model)

        return run_models

    def runs_yaml_exists(self) -> bool:
        """Check if runs.yaml exists."""
        if self.workspace.runs_yaml.exists():
            op_queue.add_no_op(
                description='Not creating runs.yaml',
                extra_description=f'{self.workspace.runs_yaml} exists')
            return True
        return False

    def data_locations_exist(self, models: Sequence[Run]) -> bool:
        """Check IDS coordinates and raise if any exist."""
        any_exists = False

        for model in models:
            if model.data_in.exists():
                logger.info('Target %s already exists', model.data_in)
                op_queue.add_no_op(
                    description='Not creating IDS',
                    extra_description=f'IDS entry {model.data_in} exists')
                any_exists = True

        return any_exists

    def run_dirs_exist(self, models: Sequence[Run]) -> bool:
        """Check if any of the run dirs exist."""
        any_exists = False

        for model in models:
            run_drc = self.workspace.cwd / model.dirname

            if run_drc.exists():
                op_queue.add_no_op(
                    description='Not creating directory',
                    extra_description=f'Directory {run_drc} exists',
                )

            any_exists = True

        return any_exists

    def warn_no_create_runs(self):
        """Add warning to op queue if runs will not be created."""
        op_queue.add_no_op(description='Not creating runs',
                           extra_description='some targets already exist, '
                           'use --force to override')

    @add_to_op_queue('Setting inital condition of', '{data_in}', quiet=True)
    def apply_operations(self, data_in: ImasHandle, run_dir: Path,
                         operations: List[Any]):
        for model in operations:
            apply_model(model, run_dir=run_dir, ids_mapping=data_in)

    @add_to_op_queue('Writing runs', '{self.workspace.runs_yaml}', quiet=True)
    def write_runs_file(self, runs: Sequence[Run]) -> None:
        runs = Runs.parse_obj(runs)
        with open(self.workspace.runs_yaml, 'w') as f:
            runs.yaml(stream=f)

    @add_to_op_queue('Writing csv', quiet=True)
    def write_runs_csv(self, runs: Sequence[Run], fname: str = 'data.csv'):
        run_map = {run.dirname: run.data_out.dict() for run in runs}
        df = pd.DataFrame.from_dict(run_map, orient='index')
        df.to_csv(fname)

    def create_run(self, model: Run, *, force: bool = False):
        """Take a run model and create it."""
        run_drc = self.workspace.cwd / model.dirname

        op_queue.add(action=run_drc.mkdir,
                     kwargs={
                         'parents': True,
                         'exist_ok': force
                     },
                     description='Creating run',
                     extra_description=f'{run_drc}')

        self.source.copy_data_to(model.data_in)

        self.system.copy_from_template(self.template_drc, run_drc)

        self.apply_operations(model.data_in, run_drc, model.operations)

        self.system.write_batchfile(self.workspace, model.dirname,
                                    self.template_drc)

        self.system.update_imas_locations(run=run_drc,
                                          inp=model.data_in,
                                          out=model.data_out)


def create(*, force, **kwargs):
    """Create input for jetto and IDS data structures.

    Parameters
    ----------
    force : bool
        Override protection if data and directories already exist.

    **kwargs
        Unused.
    """
    create_mgr = CreateManager()

    ops_list = create_mgr.generate_ops_list()

    runs = create_mgr.make_run_models(ops_list)

    if not force:

        target_exists = any([
            create_mgr.runs_yaml_exists(),
            create_mgr.data_locations_exist(runs),
            create_mgr.run_dirs_exist(runs),
        ])

        if target_exists:
            create_mgr.warn_no_create_runs()
            return

    create_mgr.write_runs_file(runs)
    create_mgr.write_runs_csv(runs)

    for model in runs:
        create_mgr.create_run(model, force=force)


def recreate(*, runs, **kwargs):
    """Create input for jetto and IDS data structures.

    Parameters
    ----------
    runs : Sequence[str]
        Run names to recreate.
    **kwargs
        Unused.
    """
    create_mgr = CreateManager()

    run_dict = {str(run.dirname): run for run in create_mgr.workspace.runs}

    run_models = []
    for run in runs:
        if run not in run_dict:
            raise ValueError(f'`{run}` not in `runs.yaml`.')

        model = run_dict[run]
        model.data_in = ImasHandle.parse_obj(model.data_in)
        model.data_out = ImasHandle.parse_obj(model.data_out)

        model.data_in.queue_delete()
        remove_run(model)

        run_models.append(model)

    for model in run_models:
        create_mgr.create_run(model)



import logging
from typing import List, Sequence, Union

import numpy as np
import xarray as xr

logger = logging.getLogger(__name__)


def standardize_time(ds: xr.Dataset, *, start: int = 0) -> None:
    """Standardize the time within a dataset.

    Simply subtracts time[0] from all time entries and adds `start`
    Note: this does not interpolate the times between different datasets

    Parameters
    ----------
    ds : xr.Dataset
        Source dataset
    start : int, optional
        Where to start the returned time series
    """

    ds['time'] = ds['time'] - ds['time'][0] + start


def standardize_grid(ds: xr.Dataset,
                     *,
                     new_dim: str,
                     old_dim: str,
                     group: str = None,
                     new_dim_data: Union[np.ndarray, int] = 0) -> xr.Dataset:
    """Standardize the grid within a dataset.

    Perform `split-apply-combine` routine on the data. Split
    by the `group`, standardize the data in `new_dim` using
    `new_dim_data` (interpolate if necessary),
    and combine replacing `old_dim` by `new_dim`.

    Parameters
    ----------
    ds : xr.Dataset
        Source dataset
    new_dim : str
        Must be an existing variable with `group` as a dimension.
    old_dim : str
        Must be an existing dimension without coordinates.
    group : str, optional
        Split the data in groups over this dimension.
    new_dim_data : Union[np.ndarray, int], optional
        The data to be used for `new_dim`. If it is an integer,
        use it as an index to grab the data from `new_dim`.

    Returns
    -------
    xr.Dataset
        New dataset with `new_dim` as a coordinate dimension.
    """
    if isinstance(new_dim_data, int):
        new_dim_data = ds.isel(  # type: ignore
            **{group: new_dim_data})[new_dim].data  # type:ignore

    gb = ds.groupby(group)

    interp_kwargs = {new_dim: new_dim_data}

    def standardize(group):
        group = group.swap_dims({old_dim: new_dim})
        group = group.interp(**interp_kwargs)
        return group

    return gb.map(standardize)


def rebase_on_grid(ds: xr.Dataset, *, coord_dim: str,
                   new_coords: np.ndarray) -> xr.Dataset:
    """Rebase (interpolate) the coordinate dimension to the new coordinates.

    Thin wrapper around `xarray.Dataset.interp`.

    Parameters
    ----------
    ds : xr.Dataset
        Source dataset
    coord_dim : str
        Name of the grid dimension (i.e. grid variable).
    new_coords : np.ndarray
        The coordinates to interpolate to

    Returns
    -------
    xr.Dataset
        Rebased dataset
    """
    return ds.interp(coords={coord_dim: new_coords},
                     kwargs={'fill_value': 'extrapolate'})


def rebase_on_time(ds: xr.Dataset,
                   *,
                   time_dim='time',
                   new_coords: np.ndarray) -> xr.Dataset:
    """Rebase (interpolate) the time dimension to the new coordinates.

    Thin wrapper around `xarray.Dataset.interp`.

    Parameters
    ----------
    ds : xr.Dataset
        Source dataset
    time_dim : str
        Name of the time dimension (i.e. time variable).
    new_coords : np.ndarray
        The coordinates to interpolate to

    Returns
    -------
    xr.Dataset
        Rebased dataset
    """
    return rebase_on_grid(ds, coord_dim=time_dim, new_coords=new_coords)


def standardize_datasets(
    datasets: Sequence[xr.Dataset],
    *,
    grid_var: str = 'rho_tor_norm',
    grid_placeholder: str = 'x',
    time_var: str = 'time',
    reference_grid_idx: int = 0,
    reference_dataset: int = 0,
) -> List[xr.Dataset]:
    """Standardize list of datasets by applying standard rebase operations.

    Applies, in sequence:
    1. `standardize_grid`
    2. `rebase_on_grid`
    3. `rebase_on_time`

    Parameters
    ----------
    datasets : Sequence[xr.Dataset]
        List of source datasets
    grid_var : str, optional
        Name of the grid dimension (i.e. grid variable)
    grid_placeholder : str, optional
        Name of the placeholder for the grid dimension
    time_var : str, optional
        Name of the time dimension (i.e. time variable)
    reference_grid_idx : int, optional
        For each dataset, the data of the grid with this index along the time dimension
        will be used as the new coordinate for the grid dimension. Data variables will
        be rebased onto these coordinates if necessary.

        Maps to `new_dim_data` in `standardize_grid`.
    reference_dataset : int, optional
        The dataset with this index will be used as the reference for rebasing.
        The grid and time coordinates of the other datasets will be rebased
        to the reference.

    Returns
    -------
    List[xr.Dataset]
        List of output datasets
    """
    datasets = [
        standardize_grid(
            ds,
            new_dim=grid_var,
            old_dim=grid_placeholder,
            new_dim_data=reference_grid_idx,
            group=time_var,
        ) for ds in datasets
    ]

    reference_grid = datasets[reference_dataset][grid_var].data

    datasets = [
        rebase_on_grid(ds, coord_dim=grid_var, new_coords=reference_grid)
        for ds in datasets
    ]

    reference_time = datasets[reference_dataset][time_var].data

    datasets = [
        rebase_on_time(ds, time_dim=time_var, new_coords=reference_time)
        for ds in datasets
    ]

    return datasets



import logging
from pathlib import Path

import click

from ..config import cfg

logger = logging.getLogger(__name__)
info, debug = logger.info, logger.debug


class Job:

    def __init__(self, dir: Path):
        self.dir = Path(dir)

    def __repr__(self):
        run = str(self.dir)
        return f'{self.__class__.__name__}({run!r})'

    @property
    def has_submit_script(self) -> bool:
        return self.submit_script.exists()

    @property
    def has_status(self) -> bool:
        return self.status_file.exists()

    @property
    def is_submitted(self) -> bool:
        return (self.dir / 'duqtools.submit.lock').exists()

    def status_file_contains(self, msg) -> bool:
        sf = self.status_file
        with open(sf) as f:
            content = f.read()
            debug('Checking if content of %s file: %s contains %s', sf,
                  content, msg)
            return msg in content

    @property
    def is_completed(self) -> bool:
        return self.status_file_contains(cfg.status.msg_completed)

    @property
    def is_failed(self) -> bool:
        return self.status_file_contains(cfg.status.msg_failed)

    @property
    def is_running(self) -> bool:
        return self.status_file_contains(cfg.status.msg_running)

    @property
    def in_file(self) -> Path:
        return self.dir / cfg.status.in_file

    @property
    def out_file(self) -> Path:
        return self.dir / cfg.status.out_file

    @property
    def status_file(self) -> Path:
        return self.dir / cfg.status.status_file

    @property
    def submit_script(self) -> Path:
        return self.dir / cfg.submit.submit_script_name

    @property
    def lockfile(self) -> Path:
        return self.dir / 'duqtools.submit.lock'

    def submit(self):
        from ..system import get_system
        debug(f'Put lockfile in place for {self.lockfile}')
        self.lockfile.touch()

        get_system().submit_job(self)

    def start(self):
        click.echo(f'Submitting {self}\033[K')
        self.submit()

        while not self.has_status:
            yield

        while self.is_running:
            yield



import logging

import click


class TermEscapeCodeFormatter(logging.Formatter):
    """A class to strip the escape codes from the."""

    def format(self, record):
        record.msg = click.unstyle(record.msg)
        return super().format(record)


class LoggingContext:
    """Context manager to Temporarily change logging configuration.

    From https://docs.python.org/3/howto/logging-cookbook.html

    Parameters
    ----------
    logger : None, optional
        Logging instance to change, defaults to root logger.
    level : None, optional
        New log level, i.e. `logging.CRITICAL`.
    handler : None, optional
        Log handler to use.
    close : bool, optional
        Whether to close the handler after use.
    """

    def __init__(self, logger=None, level=None, handler=None, close=True):
        if not logger:
            logger = logging.getLogger()
        self.logger = logger
        self.level = level
        self.handler = handler
        self.close = close

    def __enter__(self):
        if self.level is not None:
            self.old_level = self.logger.level
            self.logger.setLevel(self.level)
        if self.handler:
            self.logger.addHandler(self.handler)

    def __exit__(self, et, ev, tb):
        if self.level is not None:
            self.logger.setLevel(self.old_level)
        if self.handler:
            self.logger.removeHandler(self.handler)
        if self.handler and self.close:
            self.handler.close()


# Logger to use in duqtools
duqlog_screen = logging.getLogger(
    'screen')  # logger to log to the screen (and the log)


def initialize_duqlog_screen():
    # Logger for stdout
    stream = logging.StreamHandler()
    stream.setFormatter(logging.Formatter('%(message)s'))
    duqlog_screen.addHandler(stream)


initialize_duqlog_screen()



from typing import Dict, Sequence

import xarray as xr

from ..config import var_lookup
from ..operations import add_to_op_queue
from ..schema import IDSVariableModel
from ._handle import ImasHandle
from ._rebase import rebase_on_time, standardize_grid


def _get_placeholder_dim(grid_var: IDSVariableModel,
                         time_var: IDSVariableModel) -> str:
    dims = [dim for dim in grid_var.dims if dim != time_var.name]

    if len(dims) != 1:
        raise ValueError('Number of non-time dimensions must be 1.')

    return dims[0]


def raise_if_ids_inconsistent(*variables: IDSVariableModel):
    idss = {var.ids for var in variables}
    if len(idss) != 1:
        raise ValueError('Variables must have the same IDS, '
                         f'got {len(idss)} different values: {idss}')


@add_to_op_queue('Merge', '{target}')
def queue_merge_data(*args, **kwargs):
    """Queued version of `merge_data()`."""
    return merge_data(*args, **kwargs)


def merge_data(
    source_data: Dict[str, ImasHandle],
    target: ImasHandle,
    time_var: str,
    grid_var: str,
    data_vars: Sequence[str],
):
    time_var = var_lookup[time_var]
    grid_var = var_lookup[grid_var]
    data_vars = [var_lookup[data_var] for data_var in data_vars]

    raise_if_ids_inconsistent(time_var, grid_var, *data_vars)

    TIME_DIM = time_var.name
    GRID_DIM = grid_var.name
    PLACEHOLDER_DIM = _get_placeholder_dim(grid_var, time_var)
    RUN_DIM = 'run'
    ids = grid_var.ids

    variables = [time_var, grid_var, *data_vars]

    target_data_map = target.get(ids)

    # pick first time step as basis
    GRID_DIM = grid_var.name

    grid_dim_data = target_data_map.get_at_index(grid_var, 0)
    time_dim_data = target_data_map[time_var.path]

    datasets = []
    for run, handle in source_data.items():
        ds = handle.get_variables(variables=variables)

        ds = standardize_grid(
            ds,
            new_dim=GRID_DIM,
            old_dim=PLACEHOLDER_DIM,
            new_dim_data=grid_dim_data,
            group=TIME_DIM,
        )

        datasets.append(ds)

    datasets = [
        rebase_on_time(ds, time_dim=TIME_DIM, new_coords=time_dim_data)
        for ds in datasets
    ]

    run_data = xr.concat(datasets, RUN_DIM)

    means = run_data.mean(dim=RUN_DIM)
    stdevs = run_data.std(dim=RUN_DIM)

    for var in data_vars:
        if var.name not in run_data.data_vars:
            continue

        for i, time in enumerate(run_data.time):
            mean = means.isel({TIME_DIM: i})[var.name].data
            stdev = stdevs.isel({TIME_DIM: i})[var.name].data

            target_data_map.set_at_index(var.path, value=mean, index=i)
            target_data_map.set_at_index(var.path + '_error_upper',
                                         value=mean + stdev,
                                         index=i)

    target_data_map.sync(target)



import functools
import logging
from datetime import datetime
from pathlib import Path
from sys import exit, stderr, stdout

import click
from pydantic import ValidationError

from ._logging_utils import TermEscapeCodeFormatter, duqlog_screen
from .config import cfg
from .operations import op_queue, op_queue_context

logger = logging.getLogger(__name__)

try:
    import coverage
    coverage.process_startup()
except ImportError:
    pass


def config_option(f):
    return click.option('-c',
                        '--config',
                        default='duqtools.yaml',
                        help='Path to config.')(f)


def quiet_option(f):
    return click.option('-q',
                        '--quiet',
                        is_flag=True,
                        default=False,
                        help='Don\'t output anything to the screen'
                        ' (except mandatory prompts).')(f)


def debug_option(f):
    return click.option('--debug',
                        is_flag=True,
                        help='Enable debug print statements.')(f)


def dry_run_option(f):
    return click.option('--dry-run',
                        is_flag=True,
                        help='Execute without any side-effects.')(f)


def yes_option(f):
    return click.option('--yes',
                        is_flag=True,
                        help='Answer yes to questions automatically.')(f)


def logfile_option(f):
    return click.option('--logfile',
                        '-l',
                        is_flag=False,
                        default='duqtools.log',
                        help='where to send the logfile,'
                        ' the special values stderr/stdout'
                        ' will send it there respectively.')(f)


def parse_common_options(func):
    """With this function it becomes possible to parse the options in user
    defined order."""

    def parse_options(ctx, *, logfile, debug, quiet, dry_run, config, yes,
                      **kwargs):
        # Config option
        if ctx.command.name != 'init':
            try:
                cfg.parse_file(config)
            except ValidationError as e:
                exit(e)

        # Debug option
        if debug:
            level = logging.DEBUG
        else:
            level = logging.INFO

        # Logfile option
        streams = {'stdout': stdout, 'stderr': stderr}
        logging.getLogger().handlers = []

        if logfile in streams.keys():
            logging.basicConfig(stream=streams[logfile], level=level)
        else:
            fhandler = logging.FileHandler(logfile)
            fhandler.setLevel(level)

            # Remove fancies from logfiles
            escaped_format = TermEscapeCodeFormatter(logging.BASIC_FORMAT)
            fhandler.setFormatter(escaped_format)
            logging.getLogger().addHandler(fhandler)

        logger.info('')
        logger.info(
            'Duqtools starting at '
            f'{datetime.now().astimezone().strftime("%Y-%m-%d %H:%M:%S %z")}')
        logger.info('------------------------------------------------')
        logger.info('')

        # Yes option
        op_queue.yes = yes

        # Dry run option
        op_queue.dry_run = dry_run

        # Quiet option
        if quiet:
            duqlog_screen.handlers = []  # remove output methods
            cfg.quiet = True

    def callback(ctx, **kwargs):
        parse_options(ctx, **kwargs)
        func(**kwargs)

    return callback


def common_options(func):
    """common_options.

    IMPORTANT: This must be the last click option specified before
    the execution of the actual function, otherwise options will be
    missing
    """
    wrapper = click.pass_context(parse_common_options(func))

    for option in (logfile_option, debug_option, config_option, quiet_option,
                   dry_run_option, yes_option):
        wrapper = option(wrapper)

    functools.update_wrapper(wrapper, func)

    return wrapper


def cli_entry():
    from duqtools import fix_dependencies
    fix_dependencies()
    cli()


@click.group()
def cli(**kwargs):
    """For more information, check out the documentation:

    https://duqtools.readthedocs.io
    """
    pass


@cli.command('init')
@click.option('--force', is_flag=True, help='Overwrite existing config.')
@common_options
def cli_init(**kwargs):
    """Create a default config file."""
    from .init import init
    with op_queue_context():
        try:
            init(**kwargs)
        except RuntimeError as e:
            exit(e)


@cli.command('create')
@click.option('--force',
              is_flag=True,
              help='Overwrite existing run directories and IDS data.')
@common_options
def cli_create(**kwargs):
    """Create the UQ run files."""
    from .create import create
    with op_queue_context():
        create(**kwargs)


@cli.command('recreate')
@click.argument('runs', nargs=-1)
@common_options
def cli_recreate(**kwargs):
    """Read `runs.yaml` and re-create the given runs.

    Example:
    \b
    - duqtools recreate run_0003 run_0004 --force
    """
    from .create import recreate
    with op_queue_context():
        recreate(**kwargs)


@cli.command('submit')
@click.option('--force',
              is_flag=True,
              help='Re-submit running or completed jobs.')
@click.option(
    '--schedule',
    is_flag=True,
    help='Schedule and submit jobs automatically. `max_jobs` must be defined.')
@click.option('-j',
              '--max_jobs',
              type=int,
              help='Maximum number of jobs to submit.')
@click.option('-a', '--array', is_flag=True, help='Submit jobs as array.')
@click.option('-r',
              '--resubmit',
              multiple=True,
              default=tuple(),
              type=Path,
              help='Case to re-submit, can be specified multiple times')
@common_options
def cli_submit(**kwargs):
    """Submit the UQ runs.

    This subcommand will read `runs.yaml`, and start all runs which are not yet
    running. By default, It will not re-submit running or finished jobs.

    There is a scheduler that will continue to submit jobs until the specified
    maximum number of jobs is running. Once a job has completed, a new job will
    be submitted from the queue to fill the spot.
    """
    from .submit import submit
    with op_queue_context():
        submit(**kwargs)


@cli.command('status')
@click.option('--detailed', is_flag=True, help='Detailed info on progress')
@click.option('--progress', is_flag=True, help='Fancy progress bar')
@common_options
def cli_status(**kwargs):
    """Print the status of the UQ runs."""
    from .status import status
    status(**kwargs)


@cli.command('plot')
@click.option('-x',
              'grid_var',
              default='rho_tor_norm',
              type=str,
              help='Name of the x (grid) variable')
@click.option('-y',
              'data_vars',
              type=str,
              help='Name of the y (data) variable(s)',
              multiple=True)
@click.option('-t',
              'time_var',
              type=str,
              help='Name of the time variable',
              default='time')
@click.option('-m',
              '--imas',
              'imas_paths',
              type=str,
              help='IMAS path formatted as <user>/<db>/<shot>/<number>.',
              multiple=True)
@click.option('-u', '--user', 'user', type=str, help='IMAS user.')
@click.option('-d', '--db', 'db', type=str, help='IMAS database.')
@click.option('-s', '--shot', 'shot', type=int, help='IMAS shot.')
@click.option('-r',
              '--runs',
              'runs',
              type=int,
              multiple=True,
              help='IMAS run (multiple runs can be specified)')
@click.option('-i',
              '--input',
              'input_files',
              type=str,
              help='Input file, i.e. `data.csv` or `runs.yaml`',
              multiple=True)
@click.option('-o',
              '--format',
              'extensions',
              type=str,
              help='Output format (json, html, png, svg, pdf), default: html.',
              default=('html', ),
              multiple=True)
def cli_plot(**kwargs):
    """Plot some IDS data.

    \b
    Examples:
    - duqtools plot -x rho_tor_norm -y t_i_ave -y zeff -i data.csv
    - duqtools plot -y t_i_ave -y zeff -m jet/91234/5
    - duqtools plot -y zeff -u user -d jet -s 91234 -r 5 -r 6 -r 7
    - duqtools plot -y zeff -m user/91234/5 -i data.csv
    - duqtools plot -y zeff -m user/91234/5 -o json
    """
    from .plot import plot
    plot(**kwargs)


@cli.command('clean')
@click.option('--out', is_flag=True, help='Remove output data.')
@click.option('--force',
              is_flag=True,
              help='Overwrite backup file if necessary.')
@common_options
def cli_clean(**kwargs):
    """Delete generated IDS data and the run dir."""
    from .cleanup import cleanup
    with op_queue_context():
        cleanup(**kwargs)


@cli.command('go')
@click.option('--force', is_flag=True, help='Overwrite files when necessary.')
@common_options
def cli_go(**kwargs):
    """Run create, submit, status, dash in succession.

    Useful for existing tested and working pipelines.
    """
    from .create import create
    from .dash import dash
    from .status import status
    from .submit import submit
    with op_queue_context():
        create(**kwargs)
    with op_queue_context():
        submit(max_jobs=None,
               array=True,
               schedule=None,
               resubmit=tuple(),
               **kwargs)

    skwargs = kwargs.copy()
    skwargs['detailed'] = True
    skwargs['progress'] = False
    status(**skwargs)

    dash(**kwargs)


@cli.command('yolo')
@click.pass_context
def cli_yolo(ctx, **kwargs):
    """Live on the edge, run `duqtools go --force --yes --quiet`."""
    ctx.invoke(cli_go, force=True, quiet=True, yes=True)


@cli.command('dash')
@common_options
def cli_dash(**kwargs):
    """Open dashboard for evaluating IDS data."""
    from .dash import dash
    dash(**kwargs)


@cli.command('merge')
@common_options
def cli_merge(**kwargs):
    """Merge data sets with error propagation."""
    from .merge import merge
    with op_queue_context():
        merge(**kwargs)


@cli.command('list-variables')
@config_option
def cli_list_variables(**kwargs):
    """List available variables.

    Picks up variables from `duqtools.yaml` if it exists in the local
    directory.
    """
    from .list_variables import list_variables
    list_variables(**kwargs)


if __name__ == '__main__':
    cli()



import logging

from ._handle import ImasHandle
from ._imas import imas_mocked
from ._mapping import IDSMapping
from ._merge import merge_data, queue_merge_data
from ._rebase import (rebase_on_grid, rebase_on_time, standardize_datasets,
                      standardize_grid, standardize_time)

logger = logging.getLogger(__name__)

__all__ = [
    'IDSMapping',
    'ImasHandle',
    'merge_data',
    'rebase_on_grid',
    'rebase_on_time',
    'standardize_datasets',
    'standardize_grid',
    'standardize_time',
    'imas_mocked',
    'queue_merge_data',
]



"""Function to create llcmd file."""
from __future__ import annotations

import stat
from typing import TYPE_CHECKING

import jetto_tools

if TYPE_CHECKING:
    from duqtools.models import WorkDirectory


def write_batchfile(workspace: WorkDirectory, run_name: str,
                    jset: jetto_tools.jset.JSET):
    """Write batchfile (`.llcmd`) to start jetto.

    Parameters
    ----------
    target_drc : Path
        Directory to place batch file into.
    """
    run_drc = workspace.cwd / run_name
    llcmd_path = run_drc / '.llcmd'

    full_path = workspace.cwd / run_name
    rjettov_path = full_path / 'rjettov'
    rel_path = workspace.subdir / run_name

    build_name = jset['JobProcessingPanel.name']
    build_user_name = jset['JobProcessingPanel.userid']
    machine_number = jset['JobProcessingPanel.machineNumber']
    num_proc = jset['JobProcessingPanel.numProcessors']
    wall_time = jset['JobProcessingPanel.wallTime']

    with open(llcmd_path, 'w') as f:
        f.write(f"""#!/bin/sh
#SBATCH -J duqtools.jetto.{run_name}
#SBATCH -i /dev/null
#SBATCH -o {full_path}/ll.out
#SBATCH -e {full_path}/ll.err
#SBATCH -p {machine_number}

#SBATCH -N 1
#SBATCH -n {num_proc}
#SBATCH -t {wall_time}:00:00

cd {full_path}
{rjettov_path} -S -I -p -xmpi -x64 {rel_path} \
{build_name} {build_user_name}""")

    llcmd_path.chmod(llcmd_path.stat().st_mode | stat.S_IXUSR)



import numpy as np
from pydantic import Field

from ._basemodel import BaseModel


class LinSpace(BaseModel):
    """Generated evenly spaced numbers over a specified interval.

    See the implementation of [numpy.linspace][] for more details.
    """
    start: float = Field(description='Start value of the sequence.')
    stop: float = Field(description='End value of the sequence.')
    num: int = Field(description='Number of samples to generate.')

    @property
    def values(self):
        """Convert to list."""
        # `val.item()` converts to native python types
        return [
            val.item() for val in np.linspace(self.start, self.stop, self.num)
        ]


class ARange(BaseModel):
    """Generate evenly spaced numbers within a given interval.

    See the implementation of [numpy.arange][] for more details.
    """

    start: float = Field(
        description='Start of the interval. Includes this value.')
    stop: float = Field(
        description='End of the interval. Excludes this interval.')
    step: float = Field(description='Spacing between values.')

    @property
    def values(self):
        """Convert to list."""
        # `val.item()` converts to native python types
        return [
            val.item() for val in np.arange(self.start, self.stop, self.step)
        ]



from __future__ import annotations

from pathlib import Path
from typing import List, Union

from pydantic import Field

from ._basemodel import BaseModel
from ._dimensions import IDSOperation, JettoOperation
from ._imas import ImasBaseModel


class Run(BaseModel):
    dirname: Path = Field(description='Directory of run')
    data_in: ImasBaseModel
    data_out: ImasBaseModel
    operations: List[Union[IDSOperation, JettoOperation,
                           List[Union[IDSOperation, JettoOperation]]]]


class Runs(BaseModel):
    __root__: List[Run] = []

    def __iter__(self):
        yield from self.__root__

    def __getitem__(self, index: int):
        return self.__root__[index]

    def construct(run_list: List[dict]):
        __root__ = []
        for run in run_list:
            __root__.append(Run.construct(**run))
        return __root__



from __future__ import annotations

import atexit
import logging
from collections import deque
from contextlib import contextmanager
from inspect import signature
from typing import Callable, Optional, Sequence

import click
from pydantic import Field, validator

from ._logging_utils import duqlog_screen
from .config import cfg
from .schema import BaseModel

logger = logging.getLogger(__name__)

OP_STYLE = {'fg': 'green'}

NO_OP_STYLE = {
    'fg': 'red',
    'bold': True,
}

HEADER_STYLE = {
    'fg': 'red',
    'bold': True,
}


class Operation(BaseModel):
    """Operation, simple class which has a callable action.

    Usually not called directly but used through Operations. has the
    following members:
    """

    quiet: bool = Field(False,
                        description='print out this operation to the screen')
    description: str = Field(
        description='description of the operation to be done')
    action: Optional[Callable] = Field(
        description='a function which can be executed when we '
        'decide to apply this operation')
    extra_description: Optional[str] = Field(description='Extra description')
    args: Optional[Sequence] = Field(
        None,
        description='positional arguments that have to be '
        'passed to the action')
    kwargs: Optional[dict] = Field(
        None,
        description='keyword arguments that will be '
        'passed to the action')

    def __call__(self) -> Operation:
        """Execute the action with the args and kwargs.

        Returns
        -------
        Operation
            The operation that was executed
        """
        if self.action:
            logger.debug(self.long_description)
            self.action(*self.args, **self.kwargs)  # type: ignore
        return self

    @property
    def long_description(self):
        style = NO_OP_STYLE if self.action is None else OP_STYLE

        description = click.style(self.description, **style)

        if self.extra_description is not None:
            description = f'{description} : {self.extra_description}'

        return description

    @validator('args', always=True)
    def validate_args(cls, v):
        if v is None:
            v = ()
        return v

    @validator('kwargs', always=True)
    def validate_kwargs(cls, v):
        if v is None:
            v = {}
        return v


class Operations(deque):
    """Operations Queue which keeps track of all the operations that need to be
    done.

    It's basically dask_delayed, but custom made and a few drawbacks:
    The return value from an action is eventually discarded,
    communication between queue items is possible through references, or
    global values, but not really recommended, and no guidance for this
    is provided
    """

    _instance = None
    yes = False  # Apply operations without prompt
    enabled = False  # Actually do something
    dry_run = False  # Never apply any operations (do not even ask)

    def __new__(cls, *args, **kwargs):
        # Make it a singleton
        if not Operations._instance:
            Operations._instance = super().__new__(cls)
        return Operations._instance

    def add(self, **kwargs) -> None:
        """Convenience Operation wrapper around .append().

        ```python
        from duqtools.operations import add_to_op_queue.

        op_queue.add(action=print, args=('Hello World,),
                description="Function that prints hello world")
        ```
        """
        self.append(Operation(**kwargs))

    def add_no_op(self, description: str, extra_description: str = None):
        """Adds a line to specify an action will not be undertaken."""
        self.add(action=None,
                 description=description,
                 extra_description=extra_description)

    def put(self, item: Operation):
        """synonym for append."""
        self.append(item)

    def append(self, item: Operation):  # type: ignore
        """Restrict our diet to Operation objects only."""
        if self.enabled:
            logger.debug(
                f'Appended {item.description} to the operations queue')
            super().append(item)
        else:
            duqlog_screen.info('- ' + item.long_description)
            item()

    def apply(self) -> Operation:
        """Apply the next operation in the queue and remove it."""

        operation = self.popleft()
        operation()
        return operation

    def apply_all(self) -> None:
        """Apply all queued operations and empty the queue.

        and show a fancy progress bar while applying
        """
        from tqdm import tqdm
        duqlog_screen.info(click.style('Applying Operations', **HEADER_STYLE))
        if not cfg.quiet:
            with tqdm(total=len(self), position=1) as pbar:
                pbar.set_description('Applying operations')
                with tqdm(iterable=False, bar_format='{desc}',
                          position=0) as dbar:
                    while len(self) != 0:
                        op = self.popleft()
                        if not op.quiet:
                            dbar.set_description(op.long_description)
                        logger.info(op.long_description)
                        pbar.update()
                        op()
        else:
            while len(self) != 0:
                op = self.popleft()
                op()

    def confirm_apply_all(self) -> bool:
        """First asks the user if he wants to apply everything.

        Returns
        -------
        bool: did we apply everything or not
        """

        # To print the descriptions we need to get them
        duqlog_screen.info('')
        duqlog_screen.info(
            click.style('Operations in the Queue:', **HEADER_STYLE))
        duqlog_screen.info(
            click.style('========================', **HEADER_STYLE))
        for op in self:
            if not op.quiet:
                duqlog_screen.info('- ' + op.long_description)

        if self.dry_run:
            duqlog_screen.info('Dry run enabled, not applying op_queue')
            return False

        # Do not confirm if all actions are no-op
        if all(op.action is None for op in self):
            duqlog_screen.info('\nNo actions to execute.')
            return False

        ans = self.yes or click.confirm(
            '\nDo you want to apply all these operations?', default=False)

        if ans:
            self.apply_all()

        return ans

    def check_unconfirmed_operations(self):
        """Safety check, it should never happen that operations are not
        executed."""
        if len(self) != 0:
            duqlog_screen.warning(
                click.style((f'There are still {len(self)} operations '
                             'in the queue at program exit!'), **HEADER_STYLE))


op_queue = Operations()


def confirm_operations(func):
    """Decorator which confirms and applies queued operations after the
    function.

    ```python
    from duqtools.operations import confirm_operations, op_queue

    @confirm_operations
    def complicated_stuff()
        op_queue.add(action=print, args=('Hello World,),
                description="Function that prints hello world")
        op_queue.add(action=print, args=('Hello World again,),
                description="Function that prints hello world, again")
    ```
    """

    def wrapper(*args, **kwargs):
        if op_queue.enabled:
            raise RuntimeError('op_queue already enabled')
        try:
            op_queue.enabled = True
            ret = func(*args, **kwargs)
            op_queue.confirm_apply_all()
            op_queue.clear()
            op_queue.enabled = False
        except Exception:
            op_queue.clear()
            op_queue.enabled = False
            raise
        return ret

    return wrapper


def add_to_op_queue(op_desc: str,
                    extra_desc: Optional[str] = None,
                    quiet=False):
    """Decorator which adds the function call to the op_queue, instead of
    executing it directly, the string can be a format string and use the
    function arguments.

    ```python
    from duqtools.operations import add_to_op_queue, op_queue

    @add_to_op_queue("Printing hello world", "{name}")
    def print_hello_world(name):
        print(f"Hello World {name}")


    print_hello_world("Snoozy")

    op_queue.confirm_apply_all()
    ```
    """

    def op_queue_real(func):

        def wrapper(*args, **kwargs) -> None:
            # For the description format we must convert args to kwargs
            sig = signature(func)
            args_to_kw = dict(zip(sig.parameters, args))
            fkwargs = kwargs.copy()
            fkwargs.update(args_to_kw)
            extra_formatted = None
            if extra_desc:
                extra_formatted = extra_desc.format(**fkwargs)
            op_formatted = op_desc.format(**fkwargs)

            # add the function to the queue
            op_queue.add(action=func,
                         args=args,
                         kwargs=kwargs,
                         description=op_formatted,
                         extra_description=extra_formatted,
                         quiet=quiet)

        return wrapper

    return op_queue_real


atexit.register(op_queue.check_unconfirmed_operations)


@contextmanager
def op_queue_context():
    """Context manager to enable the op_queue, and confirm_operations on exit
    Also disables the op_queue on exit.

    Works more or less the same as the `@confirm_operations` decorator
    """

    if op_queue.enabled:
        raise RuntimeError('op_queue already enabled')
    try:
        op_queue.enabled = True
        yield
        op_queue.confirm_apply_all()
        op_queue.clear()
        op_queue.enabled = False
    except Exception:
        op_queue.clear()
        op_queue.enabled = False
        raise



"""This module contains tools for interfacing with jetto runs."""

from ._jettovar_to_json import jettovar_to_json
from ._llcmd import write_batchfile
from ._system import JettoSystem

__all__ = [
    'jettovar_to_json',
    'write_batchfile',
    'JettoSystem',
]



from pathlib import Path
from warnings import warn

from .config import cfg
from .ids import ImasHandle
from .jetto import JettoSystem
from .models import AbstractSystem, Job, WorkDirectory
from .schema import JettoVar


class DummySystem(AbstractSystem):
    """This is a dummy system that implements the basic interfaces.

    It exists for testing purposes in absence of actual modelling
    software.
    """

    @staticmethod
    def write_batchfile(workspace: WorkDirectory, run_name: str,
                        template_drc: Path):
        pass

    @staticmethod
    def submit_job(job: Job):
        pass

    @staticmethod
    def copy_from_template(source_drc: Path, target_drc: Path):
        pass

    @staticmethod
    def imas_from_path(template_drc: Path):
        return ImasHandle(db='', shot='-1', run='-1')

    @staticmethod
    def update_imas_locations(run: Path, inp, out):
        pass

    @staticmethod
    def set_jetto_variable(run: Path, key: str, value, variable: JettoVar):
        pass


def get_system():
    """get_system.

    Get the system to do operations with TODO make it a variable, not a
    function
    """
    if (cfg.system in ['jetto-pythontools', 'jetto-duqtools']):
        warn(f"{cfg.system} system deprecated, please use 'jetto'",
             DeprecationWarning)

    if (cfg.system in ['jetto', 'jetto-duqtools', 'jetto-pythontools']):
        return JettoSystem
    elif (cfg.system == 'dummy'):
        return DummySystem
    else:
        raise NotImplementedError(f'system {cfg.system} is not implemented')



from __future__ import annotations

from pydantic import Field
from typing_extensions import Literal

from ._basemodel import BaseModel


class LHSSampler(BaseModel):
    """Select the Latin Hypercube sampler by specifying
    `method: latin-hypercube`."""

    method: Literal['latin-hypercube'] = 'latin-hypercube'

    n_samples: int = Field(3, description='Number of samples to take')


class HaltonSampler(BaseModel):
    """Select the Sobol sampler by specifying `method: halton`."""

    method: Literal['halton']

    n_samples: int = Field(3, description='Number of samples to take')


class SobolSampler(BaseModel):
    """Select the Sobol sampler by specifying `method: sobol`."""

    method: Literal['sobol', 'low-discrepancy-sequence']

    n_samples: int = Field(3, description='Number of samples to take')


class CartesianProduct(BaseModel):
    """Select the Cartesian product sampler by specifying
    `method: cartesian-product`."""

    method: Literal['cartesian-product'] = 'cartesian-product'



from pathlib import Path
from typing import List, Optional, Union

from pydantic import DirectoryPath, Field
from typing_extensions import Literal

from ._basemodel import BaseModel
from ._description_helpers import formatter as f
from ._dimensions import CoupledDim, OperationDim
from ._imas import ImasBaseModel
from .data_location import DataLocation
from .matrix_samplers import (CartesianProduct, HaltonSampler, LHSSampler,
                              SobolSampler)
from .variables import VariableConfigModel
from .workdir import WorkDirectoryModel


class CreateConfigModel(BaseModel):
    """The options of the `create` subcommand are stored in the `create` key in
    the config."""
    dimensions: List[Union[CoupledDim, OperationDim]] = Field(description=f("""
        The `dimensions` specifies the dimensions of the matrix to sample
        from. Each dimension is a compound set of operations to apply.
        From this, a matrix all possible combinations is generated.
        Essentially, it generates the
        [Cartesian product](en.wikipedia.org/wiki/Cartesian_product)
        of all operations. By specifying a different `sampler`, a subset of
        this hypercube can be efficiently sampled.
        """))

    sampler: Union[LHSSampler, HaltonSampler, SobolSampler,
                   CartesianProduct] = Field(default=CartesianProduct(),
                                             discriminator='method',
                                             description=f("""
        For efficient UQ, it may not be necessary to sample the entire matrix
        or hypercube. By default, the cartesian product is taken
        (`method: cartesian-product`). For more efficient sampling of the space,
        the following `method` choices are available:
        [`latin-hypercube`](en.wikipedia.org/wiki/Latin_hypercube_sampling),
        [`sobol`](en.wikipedia.org/wiki/Sobol_sequence),
        [`halton`](en.wikipedia.org/wiki/Halton_sequence).
        Where `n_samples` gives the number of samples to extract.
        """))

    template: DirectoryPath = Field(description=f("""
        Template directory to modify. Duqtools copies and updates the settings
        required for the specified system from this directory. This can be a
        directory with a finished run, or one just stored by JAMS (but not yet
        started). By default, duqtools extracts the input IMAS database entry
        from the settings file (e.g. jetto.in) to find the data to modify for
        the UQ runs.
        """))

    template_data: Optional[ImasBaseModel] = Field(description=f("""
        Specify the location of the template data to modify. This overrides the
        location of the data specified in settings file in the template
        directory.
        """))

    data: DataLocation = Field(description=f("""
        Where to store the in/output IDS data.
        The data key specifies the machine or imas
        database name where to store the data (`imasdb`). duqtools will write the input
        data files for UQ start with the run number given by `run_in_start_at`.
        The data generated by the UQ runs (e.g. from jetto) will be stored
        starting by the run number given by `run_out_start_at`.
        """))


class SubmitConfigModel(BaseModel):
    """The options of the `submit` subcommand are stored under the `submit` key
    in the config.

    The config describes the commands to start the UQ runs.
    """

    submit_script_name: str = Field(
        '.llcmd', description='Name of the submission script.')
    submit_command: str = Field('sbatch',
                                description='Submission command for slurm.')
    submit_system: Literal['prominence', 'slurm'] = Field(
        'slurm',
        description='System to submit jobs to [slurm (default), prominence]')


class StatusConfigModel(BaseModel):
    """The options of the `status` subcommand are stored under the `status` key
    in the config.

    These only need to be changed if the modeling software changes.
    """

    status_file: str = Field('jetto.status',
                             description='Name of the status file.')
    in_file: str = Field('jetto.in',
                         description=f("""
            Name of the modelling input file, will be used to check
            if the subprocess has started.
            """))

    out_file: str = Field('jetto.out',
                          description=f("""
            Name of the modelling output file, will be used to
            check if the software is running.
            """))

    msg_completed: str = Field('Status : Completed successfully',
                               description=f("""
            Parse `status_file` for this message to check for
            completion.
            """))

    msg_failed: str = Field('Status : Failed',
                            description=f("""
            Parse `status_file` for this message to check for
            failures.
            """))

    msg_running: str = Field('Status : Running',
                             description=f("""
            Parse `status_file` for this message to check for
            running status.
            """))


class MergeStep(BaseModel):
    """These parameters describe which paths should be merged.

    Three sets of variables need to be defined:
    - time_variable: this points to the data for the time coordinate
    - grid_variable: this points to the data for the grid variable
    - data_variables: these point to the data to be merged

    Note that all variables must be from the same IDS.

    The grid and data variables must share a common dimension. The grid variable
    will be used to rebase all data variables to a common grid.

    The time variable will be used to rebase the grid variable and the data variables
    to a common time coordinate. To denote the time index, use `/*/` in both
    the grid and data variables.

    Rebasing involves interpolation.

    Note that multiple merge steps can be specified, for example for different
    IDS.
    """
    data_variables: List[str] = Field(description=f("""
            This is a list of data variables to be merged. This means
            that the mean and error for these data over all runs are calculated
            and written back to the ouput data location.
            The paths should contain `/*/` for the time component or other dimensions.
            """))
    grid_variable: str = Field(description=f("""
            This variable points to the data for the grid coordinate. It must share a common
            placeholder dimension with the data variables.
            It will be used to rebase all data variables to same (radial) grid before merging
            using interpolation.
            The path should contain '/*/' to denote the time component or other dimension.
            """))
    time_variable: str = Field(description=f("""
            This variable determines the time coordinate to merge on. This ensures
            that the data from all runs are on the same time coordinates before
            merging.
            """))


class MergeConfigModel(BaseModel):
    """The options of the `merge` subcommand are stored under the `merge` key
    in the config.

    These keys define the location of the IMAS data, which IDS entries
    to merge, and where to store the output.

    Before merging, all keys are rebased on (1) the same radial
    coordinate specified via `base_ids` and (2) the timestamp.
    """
    data: Path = Field('runs.yaml',
                       description=f("""
            Data file with IMAS handles, such as `data.csv` or `runs.yaml`'
            """))
    template: ImasBaseModel = Field(description=f("""
            This IMAS DB entry will be used as the template.
            It is copied to the output location.
            """))
    output: ImasBaseModel = Field(
        description='Merged data will be written to this IMAS DB entry.')
    plan: List[MergeStep] = Field(description='List of merging operations.')


class ConfigModel(BaseModel):
    """The options for the CLI are defined by this model."""
    submit: SubmitConfigModel = Field(
        SubmitConfigModel(),
        description='Configuration for the submit subcommand')

    create: Optional[CreateConfigModel] = Field(
        description='Configuration for the create subcommand')

    status: StatusConfigModel = Field(
        StatusConfigModel(),
        description='Configuration for the status subcommand')

    merge: Optional[MergeConfigModel] = Field(
        description='Configuration for the merge subcommand')

    workspace: WorkDirectoryModel

    extra_variables: Optional[VariableConfigModel] = Field(
        description='Specify extra variables for this run.')

    system: Literal['jetto', 'dummy', 'jetto-pythontools',
                    'jetto-duqtools'] = Field(
                        'jetto', description='backend system to use')

    quiet: bool = Field(
        False,
        description='dont output to stdout, except for mandatory prompts')



import sys
from getpass import getuser
from pathlib import Path

import pandas as pd
import streamlit as st

from duqtools.ids import ImasHandle, merge_data
from duqtools.utils import read_imas_handles_from_file

sys.path.insert(0, str(Path(__file__).parent))

from _shared import get_var_options  # noqa
from _shared import default_workdir, get_ids_options, get_variables  # noqa

st.markdown('# Merge IMAS data')

with st.sidebar:
    st.header('Input data')
    work_dir = st.text_input('Work directory', default_workdir)
    data_file = st.text_input('Data file', 'data.csv')

inp = Path(work_dir) / data_file

if not inp.exists():
    raise ValueError(f'{inp} does not exist!')

handles = read_imas_handles_from_file(inp)
df = pd.DataFrame.from_dict(
    {index: model.dict()
     for index, model in handles.items()}, orient='index')

with st.expander('Click to show runs'):
    st.table(df)

with st.sidebar:
    ids_options = get_ids_options()

    ids = st.selectbox('Select IDS', ids_options, index=0)

    var_options = get_var_options(ids=ids)

    default_x_key = var_options.index('rho_tor_norm')
    default_y_val = 't_i_ave'

    x_key = st.selectbox('Select x', var_options, index=default_x_key)

    y_keys = st.multiselect('Select y', var_options, default=default_y_val)

variables = get_variables(ids=ids, x_key=x_key, y_keys=y_keys)

with st.form('merge_form'):

    st.subheader('Merge data')

    st.write("""
        With this form you can merge all runs into a new IMAS DB entry.

        The mean and standard deviation are calculated over all the
        runs for the fields specified in the side bar.

        Note that it does not do error propagation yet if the source data have
        error bars already.
        """)

    a_run = df.iloc[0]

    st.markdown('**Template IMAS entry**')
    st.write(
        """This IMAS entry will be used as the template. The template is copied,
        and any existing data is overwritten for the given fields.""")

    cols = st.columns((20, 20, 30, 30))

    template = {
        'user': cols[0].text_input('User',
                                   value=a_run.user,
                                   key='user_template'),
        'db': cols[1].text_input('Machine', value=a_run.db, key='db_template'),
        'shot': cols[2].number_input('Shot',
                                     value=a_run.shot,
                                     key='shot_template'),
        'run': cols[3].number_input('Run', value=a_run.run,
                                    key='run_template'),
    }

    template = ImasHandle(**template)

    st.markdown('**Target IMAS entry**')
    st.write('The data will be stored in the DB entry given below.')

    cols = st.columns((20, 20, 30, 30))

    target = {
        'user':
        cols[0].text_input('User',
                           value=getuser(),
                           key='user_target',
                           disabled=True),
        'db':
        cols[1].text_input('Machine', value=a_run.db, key='db_target'),
        'shot':
        cols[2].number_input('Shot', value=a_run.shot, key='shot_target'),
        'run':
        cols[3].number_input('Run', step=1, key='run_target'),
    }

    target = ImasHandle(**target)

    submitted = st.form_submit_button('Save')

    if submitted:
        template.copy_data_to(target)

        merge_data(source_data=handles, target=target, **variables)

        st.success('Success!')
        st.balloons()



import logging
import subprocess as sp
from time import sleep

from jetto_tools import config, template

from .config import cfg
from .jetto._system import jetto_lookup
from .models import Job, WorkDirectory

logger = logging.getLogger(__name__)
info, debug = logger.info, logger.debug
stream = logging.StreamHandler()
stream.setFormatter(logging.Formatter('%(message)s'))
logger.addHandler(stream)


class StatusError(Exception):
    ...


class Status():

    jobs: list
    jobs_submit: list
    jobs_submitted: list
    jobs_status: list
    jobs_failed: list
    jobs_running: list
    jobs_unknown: list

    def __init__(self):
        debug('Submit config: %s', cfg.submit)

        workspace = WorkDirectory.parse_obj(cfg.workspace)
        runs = workspace.runs

        self.jobs = [Job(run.dirname) for run in runs]
        debug('Case directories: %s', self.jobs)

        debug('Total number of directories: %i', len(self.jobs))

    def update_status(self):

        self.jobs_submit = [job for job in self.jobs if job.has_submit_script]
        self.jobs_status = [job for job in self.jobs if job.has_status]
        self.jobs_submitted = [job for job in self.jobs if job.is_submitted]

        self.jobs_completed = [
            job for job in self.jobs_status if job.is_completed
        ]
        self.jobs_failed = [job for job in self.jobs_status if job.is_failed]
        self.jobs_running = [job for job in self.jobs_status if job.is_running]

        self.jobs_unknown = [
            job for job in self.jobs_status
            if not (job in self.jobs_completed or job in self.jobs_failed
                    or job in self.jobs_running)
        ]

    def simple_status(self):
        """stateless status."""

        self.update_status()
        info('Total number of directories with submission script : %i',
             len(self.jobs_submit))
        info('      number of not submitted jobs                 : %i',
             (len(self.jobs_submit) - len(self.jobs_status)))
        info('Total number of directories with status     script : %i',
             len(self.jobs_status))
        info('Total number of directories with Completed status  : %i',
             len(self.jobs_completed))
        info('Total number of directories with Failed    status  : %i',
             len(self.jobs_failed))
        info('Total number of directories with Running   status  : %i',
             len(self.jobs_running))
        info('Total number of directories with Unknown   status  : %i',
             len(self.jobs_unknown))

    def progress_status(self):
        """Monitor the directory for status changes."""
        from tqdm import tqdm
        pbar_a = tqdm(total=len(self.jobs), position=0)
        pbar_a.set_description('Submitted jobs            ...')
        pbar_b = tqdm(total=len(self.jobs_submit), position=1)
        pbar_b.set_description('Running jobs              ...')
        pbar_c = tqdm(total=len(self.jobs_submit), position=2)
        pbar_c.set_description('Completed jobs            ...')
        pbar_d = tqdm(total=len(self.jobs_submit), position=3)
        pbar_d.set_description('Failed? jobs              ...')
        while len(self.jobs_completed) < len(self.jobs_submit):
            pbar_a.n = len(self.jobs_submitted)
            pbar_b.n = len(self.jobs_running)
            pbar_c.n = len(self.jobs_completed)
            pbar_d.n = len(self.jobs_failed) + len(self.jobs_unknown)
            pbar_a.refresh()
            pbar_b.refresh()
            pbar_c.refresh()
            pbar_d.refresh()
            sleep(5)
            self.update_status()

    def detailed_status(self):
        from tqdm import tqdm
        """detailed_status of all separate runs."""
        monitors = []
        for i, job in enumerate(self.jobs):
            pbar = tqdm(total=100, position=i)
            monitor = Monitor(pbar=pbar, job=job)
            monitors.append(monitor)

        while not all([monitor.finished for monitor in monitors]):
            for monitor in monitors:
                monitor.update()
            sleep(5)


class Monitor():
    """Convenience class to keep track of submissions and update progress
    bars."""

    def __init__(self, pbar, job):
        self.pbar = pbar
        self.job = job
        self.outfile = None

        jetto_template = template.from_directory(job.dir)
        jetto_template.lookup.update(jetto_lookup)
        jetto_config = config.RunConfig(jetto_template)

        self.check_kwmain_flag(jetto_config)

        infile = job.in_file
        if not infile.exists():
            debug('%s does not exist, but the job is running', infile)
            return

        self.start = jetto_config.start_time
        self.end = jetto_config.end_time
        self.time = self.start

        self.finished = False

        self.set_status()

    def check_kwmain_flag(self, jetto_config):
        """Check for NLIST2/KWMAIN in jetto.jset. If this flag is not set, the
        output in `job.out_file` does not contain the output that is grepped
        for the progress.

        More info: https://github.com/duqtools/duqtools/issues/337
        """
        msg = ('Cannot show detailed status, `nlist2.KWMAIN` flag'
               ' is not set to 1 in `{self.job.status_file}`')
        if jetto_config['kwmain'] != 1:
            raise StatusError(msg)

    def set_status(self):
        if not self.job.has_status:
            status = 'No status'
        elif not self.job.is_submitted:
            status = 'Unsubmitted '
        elif self.job.is_running:
            status = 'Running     '
        elif self.job.is_failed:
            status = 'FAILED      '
        elif self.job.is_completed:
            status = 'COMPLETED   '
        else:
            status = 'UNKNOWN'

        self.pbar.set_description(
            f'{self.job.dir.name:8s}, status: {status:12s}')
        self.pbar.refresh()

    def get_steptime(self):
        if not self.job.out_file.exists():
            debug(
                f'{self.job.out_file} does not exists, but the job is running')
            return None

        of = self.job.out_file
        cmd = f'tac {of} | grep -m 1 ^\\s*STEP'
        ret = sp.run(cmd, shell=True, capture_output=True)
        if len(ret.stdout) > 0:
            return float(ret.stdout.split('=')[2].lstrip(' ').split(' ')[0])
        return None

    def update(self):
        self.set_status()
        if self.job.has_status:
            if self.job.is_completed or self.job.is_failed:
                self.pbar.n = 100 if self.job.is_completed else 0
                self.pbar.refresh()
                self.finished = True
                return
            if not self.job.is_running:
                return

        steptime = self.get_steptime()
        if steptime:
            self.time = steptime

        self.pbar.n = int(100 * (self.time - self.start) /
                          (self.end - self.start))
        self.pbar.refresh()


def status(*, progress: bool, detailed: bool, **kwargs):
    """Show status of runs.

    Parameters
    ----------
    progress : bool
        Show progress bar.
    detailed : bool
        Show detailed progress for every job.
    """

    tracker = Status()

    if detailed:
        tracker.detailed_status()
    elif progress:
        tracker.progress_status()
    else:
        tracker.simple_status()



from __future__ import annotations

import logging
import os
import shutil
from pathlib import Path

from .config import cfg
from .ids import ImasHandle
from .models import WorkDirectory
from .operations import op_queue
from .schema.runs import Run

logger = logging.getLogger(__name__)


def remove_files(*filenames: str):
    for filename in filenames:
        try:
            os.unlink(filename)
        except FileNotFoundError:
            pass


def remove_run(model: Run):
    """Remove run directory if it exists."""
    if Path(model.dirname).exists():
        op_queue.add(
            action=shutil.rmtree,
            args=(model.dirname, ),
            description='Removing run dir',
            extra_description=f'{model.dirname}',
        )


def cleanup(out, force, **kwargs):
    """Read runs.yaml and clean the current directory.

    Parameters
    ----------
    out : bool
        Remove output IDS.
    """
    try:
        workspace = WorkDirectory.parse_obj(cfg.workspace)
        runs = workspace.runs
    except OSError:
        runs = ()
    else:
        if workspace.runs_yaml.exists and not force:
            if workspace.runs_yaml_old.exists():
                raise OSError(
                    '`runs.yaml.old` exists, use --force to overwrite anyway')

    for run in runs:
        data_in = ImasHandle.parse_obj(run.data_in)
        data_out = ImasHandle.parse_obj(run.data_out)

        data_in.queue_delete()

        if out:
            data_out.queue_delete()
        else:
            op_queue.add_no_op(description='NOT Removing',
                               extra_description=f'{data_out}')

        remove_run(run)

    op_queue.add(
        action=shutil.move,
        args=(workspace.runs_yaml, workspace.runs_yaml_old),
        description='Moving runs.yaml',
        extra_description=f'{workspace.runs_yaml_old}',
    )

    op_queue.add(
        action=remove_files,
        args=(
            'duqtools_slurm_array.err',
            'duqtools_slurm_array.out',
            'duqtools_slurm_array.sh',
        ),
        description='Removing other files',
    )



from __future__ import annotations

import logging
from getpass import getuser
from pathlib import Path
from typing import TYPE_CHECKING

from packaging import version

from .._logging_utils import LoggingContext
from ..operations import add_to_op_queue
from ._imas import Parser, imas

if TYPE_CHECKING:
    from .ids import ImasHandle


def get_imas_ual_version():
    """Get imas/ual versions.

    Parsed from a string like: `imas_3_34_0_ual_4_9_3`
    """
    vsplit = imas.names[0].split('_')

    imas_version = version.parse('.'.join(vsplit[1:4]))
    ual_version = version.parse('.'.join(vsplit[5:8]))

    return imas_version, ual_version


def add_provenance_info(handle: ImasHandle, ids: str = 'core_profiles'):
    """Add provenance information to handle.

    Parameters
    ----------
    handle : ImasHandle
        Handle to add provenance information to.
    ids : str, optional
        Which IDS to add provenance to.
    """

    import git
    import pkg_resources  # type: ignore

    with handle.open() as data_entry_target:
        entry = data_entry_target.get(ids)

        # Set the name
        entry.code.name = 'dUQtools'

        # Get the commit if we are in a repository
        try:
            entry.code.commit = git.Repo(
                Path(__file__).parent,
                search_parent_directories=True).head.object.hexsha
        except Exception:
            entry.code.commit = 'unknown'

        # Set the version if available
        try:
            entry.code.version = pkg_resources.get_distribution(
                'duqtools').version
        except Exception:
            entry.code.version = 'unknown'

        # The repository, always set to duqtools
        entry.code.repository = 'https://github.com/duqtools/duqtools/'

        entry.put(db_entry=data_entry_target)


@add_to_op_queue('Copy ids from template to', '{target}', quiet=True)
def copy_ids_entry(source: ImasHandle, target: ImasHandle):
    """Copies the ids entry to a new location.

    Parameters
    ----------
    source : ImasHandle
        Source ids entry
    target : ImasHandle
        Target ids entry

    Raises
    ------
    KeyError
        If the IDS entry you are trying to copy does not exist.
    """
    assert target.user == getuser()

    imas_version, _ = get_imas_ual_version()

    idss_in = imas.ids(source.shot, source.run)
    op = idss_in.open_env(source.user, source.db, str(imas_version.major))

    ids_not_found = op[0] < 0
    if ids_not_found:
        raise KeyError('The entry you are trying to copy does not exist')

    idss_out = imas.ids(target.shot, target.run)

    idss_out.create_env(target.user, target.db, str(imas_version.major))
    idx = idss_out.expIdx

    parser = Parser.load_idsdef()

    # Temporarily hide warnings, because this loop is very spammy
    with LoggingContext(level=logging.CRITICAL):

        for ids_info in parser.idss:
            name = ids_info['name']
            maxoccur = int(ids_info['maxoccur'])

            if name in ('ec_launchers', 'numerics', 'sdn'):
                continue

            for i in range(maxoccur + 1):
                ids = idss_in.__dict__[name]
                ids.get(i)
                ids.setExpIdx(idx)  # this line sets the index to the output
                ids.put(i)

    idss_in.close()
    idss_out.close()

    add_provenance_info(handle=target)



from __future__ import annotations

from collections import abc
from functools import singledispatch
from pathlib import Path

from .schema import BaseModel, JettoOperation


@singledispatch
def apply_model(model: BaseModel, **kwargs):
    """Apply operation in model to target. Data are modified in-place.

    Parameters
    ----------
    model
        The model describes the operation to apply to the data.

    Raises
    ------
    NotImplementedError
        When the model is unknown
    """
    raise NotImplementedError(f'Unknown model: {model}')


@apply_model.register
def _apply_coupled(model: abc.Iterable, **kwargs):  # type: ignore
    for submodel in model:
        apply_model(submodel, **kwargs)


from .ids._apply_model import _apply_ids  # noqa: E402, F401


@apply_model.register
def _apply_jetto(model: JettoOperation, run_dir: Path, **kwargs):
    from .system import get_system
    system = get_system()
    system.set_jetto_variable(run_dir, model.variable.name, model.value,
                              model.variable.lookup)



from __future__ import annotations

import logging
from abc import ABC, abstractmethod
from typing import TYPE_CHECKING

from ..schema import BaseModel, ImasBaseModel, JettoVar

if TYPE_CHECKING:
    from pathlib import Path

    from ._job import Job
    from .workdir import WorkDirectory

logger = logging.getLogger(__name__)


class AbstractSystem(ABC, BaseModel):

    @staticmethod
    @abstractmethod
    def write_batchfile(workspace: WorkDirectory, run_name: str,
                        template_drc: Path):
        pass

    @staticmethod
    @abstractmethod
    def submit_job(job: Job):
        pass

    @staticmethod
    @abstractmethod
    def copy_from_template(source_drc: Path, target_drc: Path):
        pass

    @staticmethod
    @abstractmethod
    def imas_from_path(template_drc: Path):
        pass

    @staticmethod
    @abstractmethod
    def update_imas_locations(run: Path, inp: ImasBaseModel,
                              out: ImasBaseModel):
        pass

    @staticmethod
    @abstractmethod
    def set_jetto_variable(run: Path, key: str, value, variable: JettoVar):
        pass



import logging
import shutil
import stat
import subprocess as sp
from pathlib import Path
from typing import Any, List, Sequence

from importlib_resources import files
from jetto_tools import config
from jetto_tools import job as jetto_job
from jetto_tools import jset, lookup, namelist, template

from ..config import cfg
from ..ids import ImasHandle
from ..models import AbstractSystem, Job, WorkDirectory
from ..operations import add_to_op_queue
from ..schema import JettoVar
from ._jettovar_to_json import jettovar_to_json
from ._llcmd import write_batchfile as jetto_write_batchfile

logger = logging.getLogger(__name__)

lookup_file = files('duqtools.data') / 'jetto_tools_lookup.json'
jetto_lookup = lookup.from_file(lookup_file)


class JettoSystem(AbstractSystem):

    @staticmethod
    @add_to_op_queue('Writing new batchfile', '{run_name}', quiet=True)
    def write_batchfile(workspace: WorkDirectory, run_name: str,
                        template_drc: Path):

        jetto_jset = jset.read(template_drc / 'jetto.jset')

        return jetto_write_batchfile(workspace, run_name, jetto_jset)

    @staticmethod
    def submit_job(job: Job):
        if cfg.submit.submit_system == 'slurm':
            JettoSystem.submit_slurm(job)
        elif cfg.submit.submit_system == 'prominence':
            JettoSystem.submit_prominence(job)
        else:
            raise NotImplementedError(
                'submission type {cfg.submit.submit_system}'
                ' not implemented')

    @staticmethod
    def submit_slurm(job: Job):
        if not job.has_submit_script:
            raise FileNotFoundError(job.submit_script)

        submit_cmd = cfg.submit.submit_command.split()
        cmd: List[Any] = [*submit_cmd, str(job.submit_script)]

        logger.info(f'submitting script via slurm {cmd}')

        ret = sp.run(cmd, check=True, capture_output=True)
        logger.info('submission returned: ' + str(ret.stdout))
        with open(job.lockfile, 'wb') as f:
            f.write(ret.stdout)

    @staticmethod
    def submit_prominence(job: Job):

        jetto_template = template.from_directory(job.dir)
        jetto_config = config.RunConfig(jetto_template)
        jetto_manager = jetto_job.JobManager()

        _ = jetto_manager.submit_job_to_prominence(jetto_config, job.dir)

    @staticmethod
    def submit_array(jobs: Sequence[Job]):
        if cfg.submit.submit_system == 'slurm':
            JettoSystem.submit_array_slurm(jobs)
        else:
            raise NotImplementedError(
                'array submission type {cfg.submit.submit_system}'
                ' not implemented')

    @staticmethod
    @add_to_op_queue('Submit single array job', 'duqtools_slurm_array.sh')
    def submit_array_slurm(jobs: Sequence[Job]):
        for job in jobs:
            job.lockfile.touch()

        # Get the first jobs submission script as a template
        template = []
        for line in open(jobs[0].submit_script).readlines():
            if line.startswith('#SBATCH') or line.startswith('#!'):
                template.append(line)
        # Append our own options, later options have precedence
        template.append('#SBATCH -o duqtools_slurm_array.out\n')
        template.append('#SBATCH -e duqtools_slurm_array.err\n')
        template.append('#SBATCH --array=0-' + str(len(jobs) - 1) + '\n')
        template.append('#SBATCH -J duqtools_array\n')

        scripts = [str(job.submit_script) for job in jobs]
        script_str = 'scripts=(' + ' '.join(scripts) + ')\n'
        template.append(script_str)

        template.append('echo executing ${scripts[$SLURM_ARRAY_TASK_ID]}\n')
        template.append('${scripts[$SLURM_ARRAY_TASK_ID]} || true\n')

        logger.info('writing duqtools_slurm_array.sh file')
        with open('duqtools_slurm_array.sh', 'w') as f:
            f.write(''.join(template))

        submit_cmd = cfg.submit.submit_command.split()
        cmd: List[Any] = [*submit_cmd, 'duqtools_slurm_array.sh']

        logger.info(f'Submitting script via: {cmd}')

        ret = sp.run(cmd, check=True, capture_output=True)
        logger.info('submission returned: ' + str(ret.stdout))
        with open(job.lockfile, 'wb') as f:
            f.write(ret.stdout)

    @staticmethod
    def _apply_patches_to_template(jetto_template: template.Template):
        """Apply settings that are necessary for duqtools to function."""
        # Force output of IDS data
        # https://github.com/duqtools/duqtools/issues/343
        jetto_template.jset._settings['JobProcessingPanel.selIdsRunid'] = False

    @staticmethod
    @add_to_op_queue('Copying template to', '{target_drc}', quiet=True)
    def copy_from_template(source_drc: Path, target_drc: Path):
        jetto_jset = jset.read(source_drc / 'jetto.jset')
        jetto_namelist = namelist.read(source_drc / 'jetto.in')

        jetto_sanco = None
        if (source_drc / 'jetto.sin').exists():
            jetto_sanco = namelist.read(source_drc / 'jetto.sin')

        jetto_extra = []
        if (source_drc / 'jetto.ex').exists():
            jetto_extra = [str(source_drc / 'jetto.ex')]

        jetto_template = template.Template(jset=jetto_jset,
                                           namelist=jetto_namelist,
                                           lookup=jetto_lookup,
                                           sanco_namelist=jetto_sanco,
                                           extra_files=jetto_extra)

        JettoSystem._apply_patches_to_template(jetto_template)

        jetto_config = config.RunConfig(jetto_template)

        jetto_config.export(target_drc)
        lookup.to_file(jetto_lookup, target_drc /
                       'lookup.json')  # TODO, this should be copied as well

        for filename in (
                'rjettov',
                'utils_jetto',
        ):
            src = source_drc / filename
            dst = target_drc / filename
            shutil.copyfile(src, dst)
            dst.chmod(dst.stat().st_mode | stat.S_IXUSR)

    @staticmethod
    def imas_from_path(template_drc: Path):
        jetto_jset = jset.read(template_drc / 'jetto.jset')

        return ImasHandle(
            db=jetto_jset['SetUpPanel.idsIMASDBMachine'],  # type: ignore
            user=jetto_jset['SetUpPanel.idsIMASDBUser'],  # type: ignore
            run=jetto_jset['SetUpPanel.idsIMASDBRunid'],  # type: ignore
            shot=jetto_jset['SetUpPanel.idsIMASDBShot'])  # type: ignore

    @staticmethod
    @add_to_op_queue('Updating imas locations of', '{run}', quiet=True)
    def update_imas_locations(run: Path, inp: ImasHandle, out: ImasHandle):

        jetto_template = template.from_directory(run)
        jetto_config = config.RunConfig(jetto_template)

        jetto_config['user_in'] = inp.user
        jetto_config['machine_in'] = inp.db
        jetto_config['shot_in'] = inp.shot
        jetto_config['run_in'] = inp.run

        jetto_config['machine_out'] = out.db
        jetto_config['shot_out'] = out.shot
        jetto_config['run_out'] = out.run

        jetto_config.export(run)  # Just overwrite the poor files

    @staticmethod
    def set_jetto_variable(run: Path,
                           key: str,
                           value,
                           variable: JettoVar = None):
        jetto_template = template.from_directory(run)

        if variable:
            extra_lookup = lookup.from_json(jettovar_to_json(variable))
            jetto_template._lookup.update(extra_lookup)

        jetto_config = config.RunConfig(jetto_template)

        if key == 't_start':
            jetto_config.start_time = value
        elif key == 't_end':
            jetto_config.end_time = value
        else:
            jetto_config[key] = value

        jetto_config.export(run)  # Just overwrite the poor files



from __future__ import annotations

import logging
import re
from contextlib import contextmanager
from pathlib import Path
from typing import TYPE_CHECKING, Sequence, Union

from ..config import lookup_vars
from ..operations import add_to_op_queue
from ..schema import IDSVariableModel, ImasBaseModel
from ._copy import copy_ids_entry
from ._imas import imas, imasdef
from ._mapping import IDSMapping

if TYPE_CHECKING:
    import xarray as xr

logger = logging.getLogger(__name__)

PATH_TEMPLATE = ('/afs/eufus.eu/user/g/{user}/public/imasdb/{db}'
                 '/3/0/ids_{shot}{run:04d}{suffix}')
SUFFIXES = (
    '.datafile',
    '.characteristics',
    '.tree',
)

IMAS_PATTERN = re.compile(
    r'^((?P<user>\w+)/)?(?P<db>\w+)/(?P<shot>\d+)/(?P<run>\d+)$')


def _patch_str_repr(obj: object):
    """Reset str/repr methods to default."""
    import types

    def true_repr(x):
        type_ = type(x)
        module = type_.__module__
        qualname = type_.__qualname__
        return f'<{module}.{qualname} object at {hex(id(x))}>'

    obj.__str__ = types.MethodType(true_repr, obj)  # type: ignore
    obj.__repr__ = types.MethodType(true_repr, obj)  # type: ignore


class ImasHandle(ImasBaseModel):

    @classmethod
    def from_string(cls, string: str) -> ImasHandle:
        """Return location from formatted string.

        Format:

            <user>/<db>/<shot>/<run>
            <db>/<shot>/<run>

        Default to the current user if the user is not specified.

        For example:

            g2user/jet/91234/555

        Parameters
        ----------
        string : str
            Input string containing imas db path

        Returns
        -------
        ImasHandle
        """
        match = IMAS_PATTERN.match(string)

        if match:
            return cls(**match.groupdict())

        raise ValueError(f'Could not match {string!r}')

    def to_string(self) -> str:
        """Generate string representation of Imas location."""
        return f'{self.user}/{self.db}/{self.shot}/{self.run}'

    def path(self) -> Path:
        """Return location as Path."""
        return Path(
            PATH_TEMPLATE.format(user=self.user,
                                 db=self.db,
                                 shot=self.shot,
                                 run=self.run,
                                 suffix=SUFFIXES[0]))

    def exists(self) -> bool:
        """Return true if the directory exists.

        Returns
        -------
        bool
        """
        path = self.path()
        return all(path.with_suffix(sf).exists() for sf in SUFFIXES)

    def copy_data_to(self, destination: ImasHandle):
        """Copy ids entry to given destination.

        Parameters
        ----------
        destination : ImasHandle
            Copy data to a new location.
        """
        logger.debug('Copy %s to %s', self, destination)
        try:
            copy_ids_entry(self, destination)
        except Exception as err:
            raise OSError(f'Failed to copy {self}') from err

    @add_to_op_queue('Removing ids', '{self}')
    def queue_delete(self):
        """Queued version of `ImasHandle.delete()`."""
        return self.delete()

    def delete(self):
        """Remove data from entry."""
        # ERASE_PULSE operation is yet supported by IMAS as of June 2022
        path = self.path()
        for suffix in SUFFIXES:
            to_delete = path.with_suffix(suffix)
            logger.debug('Removing %s', to_delete)
            try:
                to_delete.unlink()
            except FileNotFoundError:
                logger.warning('%s does not exist', to_delete)

    def get_raw_data(self, ids: str = 'core_profiles', **kwargs):
        """Get data from IDS entry.

        Parameters
        ----------
        ids : str, optional
            Name of profiles to open.
        **kwargs
            These keyword parameters are passed to `ImasHandle.open()`.

        Returns
        -------
        data
        """
        with self.open(**kwargs) as data_entry:
            data = data_entry.get(ids)

        # reset string representation because output is extremely lengthy
        _patch_str_repr(data)

        return data

    def get(self, ids: str = 'core_profiles') -> IDSMapping:
        """Map the data to a dict-like structure.

        Parameters
        ----------
        ids : str, optional
            Name of profiles to open

        Returns
        -------
        IDSMapping
        """
        raw_data = self.get_raw_data(ids)
        return IDSMapping(raw_data)

    def get_variables(
        self,
        variables: Sequence[Union[str, IDSVariableModel]],
        **kwargs,
    ) -> xr.Dataset:
        """Get variables from data set.

        This function looks up the data location from the
        `duqtools.config.var_lookup` table, and returns

        Parameters
        ----------
        variables : Sequence[Union[str, IDSVariableModel]]
            Variable names of the data to load.

        Returns
        -------
        ds : xarray
            The data in `xarray` format.
        **kwargs
            These keyword arguments are passed to `IDSMapping.to_xarray()`

        Raises
        ------
        ValueError
            When variables are from multiple IDSs.
        """
        var_models = lookup_vars(variables)

        idss = set(var.ids for var in var_models)

        if len(idss) > 1:
            raise ValueError(
                f'All variables must belong to the same IDS, got {idss}')

        ids = var_models[0].ids

        data_map = self.get(ids)

        ds = data_map.to_xarray(variables=var_models, **kwargs)

        return ds

    def entry(self, backend=imasdef.MDSPLUS_BACKEND):
        """Return reference to `imas.DBEntry.`

        Parameters
        ----------
        backend : optional
            Which IMAS backend to use

        Returns
        ------
        entry : `imas.DBEntry`
            IMAS database entry
        """
        return imas.DBEntry(backend, self.db, self.shot, self.run, self.user)

    @contextmanager
    def open(self, backend=imasdef.MDSPLUS_BACKEND, create: bool = False):
        """Context manager to open database entry.

        Parameters
        ----------
        backend : optional
            Which IMAS backend to use
        create : bool, optional
            Create empty database entry if it does not exist.

        Yields
        ------
        entry : `imas.DBEntry`
            Opened IMAS database entry
        """
        entry = self.entry(backend=backend)
        opcode, _ = entry.open()

        if opcode == 0:
            logger.debug('Data entry opened: %s', self)
        elif create:
            cpcode, _ = entry.create()
            if cpcode == 0:
                logger.debug('Data entry created: %s', self)
            else:
                raise OSError(
                    f'Cannot create data entry: {self}. '
                    f'Create a new db first using `imasdb {self.db}`')
        else:
            raise OSError(f'Data entry does not exist: {self}')

        try:
            yield entry
        finally:
            entry.close()



"""All public packages, functions and classes are available in this module.

Functions:

- [rebase_on_grid][duqtools.api.rebase_on_grid]
- [rebase_on_time][duqtools.api.rebase_on_time]
- [standardize_datasets][duqtools.api.standardize_datasets]

Data classes:

- [ImasHandle][duqtools.api.ImasHandle]
- [IDSMapping][duqtools.api.IDSMapping]
- [Variable][duqtools.api.Variable]

Plotting:

- [alt_errorband_chart][duqtools.api.alt_errorband_chart]
- [alt_line_chart][duqtools.api.alt_line_chart]
"""

from ._plot_utils import alt_errorband_chart, alt_line_chart
from .ids import (IDSMapping, ImasHandle, rebase_on_grid, rebase_on_time,
                  standardize_datasets)
from .schema import IDSVariableModel as Variable

__all__ = [
    'rebase_on_grid',
    'rebase_on_time',
    'standardize_datasets',
    'ImasHandle',
    'IDSMapping',
    'Variable',
    'alt_line_chart',
    'alt_errorband_chart',
]



from pathlib import Path

import pandas as pd
import streamlit as st
from _shared import (default_workdir, get_dataset, get_ids_options,
                     get_var_options, get_variables)

from duqtools._plot_utils import alt_errorband_chart, alt_line_chart
from duqtools.utils import read_imas_handles_from_file

st.title('Plot IDS')

with st.sidebar:
    st.header('Input data')
    work_dir = st.text_input('Work directory', default_workdir)
    data_file = st.text_input('Data file', 'data.csv')

inp = Path(work_dir) / data_file

if not inp.exists():
    raise ValueError(f'{inp} does not exist!')

handles = read_imas_handles_from_file(inp)
df = pd.DataFrame.from_dict(
    {index: model.dict()
     for index, model in handles.items()}, orient='index')

with st.expander('Click to show runs'):
    st.table(df)

with st.sidebar:
    ids_options = get_ids_options()

    ids = st.selectbox('Select IDS', ids_options, index=0)

    var_options = get_var_options(ids=ids)

    default_x_key = var_options.index('rho_tor_norm')
    default_y_val = 't_i_ave'

    x_key = st.selectbox('Select x', var_options, index=default_x_key)

    y_keys = st.multiselect('Select y', var_options, default=default_y_val)

    st.header('Plotting options')

    show_error_bar = st.checkbox(
        'Show errorbar',
        help=(
            'Show standard deviation band around mean y-value. All '
            'y-values are interpolated to put them on a common basis for x.'),
    )

variables = get_variables(ids=ids, x_key=x_key, y_keys=y_keys)
source = get_dataset(handles, ids=ids, **variables)

for y_key in y_keys:
    st.header(f'{x_key} vs. {y_key}')

    if show_error_bar:
        chart = alt_errorband_chart(source, x=x_key, y=y_key)
    else:
        chart = alt_line_chart(source, x=x_key, y=y_key)

    st.altair_chart(chart, use_container_width=True)



from __future__ import annotations

import logging

from .config import cfg
from .ids import ImasHandle, queue_merge_data
from .utils import read_imas_handles_from_file

logger = logging.getLogger(__name__)
info, debug = logger.info, logger.debug


def merge(**kwargs):
    """Merge data."""
    template = ImasHandle.parse_obj(cfg.merge.template)
    target = ImasHandle.parse_obj(cfg.merge.output)

    handles = read_imas_handles_from_file(cfg.merge.data)

    for step in cfg.merge.plan:
        debug('Merge input: %s', template)
        debug('Merge output: %s', target)

        template.copy_data_to(target)

        queue_merge_data(source_data=handles,
                         target=target,
                         time_var=step.time_variable,
                         grid_var=step.grid_variable,
                         data_vars=step.data_variables)



import logging
import shutil
from pathlib import Path

from importlib_resources import files

from .operations import op_queue

logger = logging.getLogger(__name__)


def init(*, config: str, force: bool, **kwargs):
    """Initialize a brand new config file with all the default values.

    Parameters
    ----------
    config : str
        Filename of the config.
    force : bool
        Overwrite config if it already exists.
    **kwargs
        Unused.

    Raises
    ------
    RuntimeError
        When the config already exists.
    """
    src = files('duqtools.data') / 'duqtools.yaml'

    config_filepath = Path(config)

    if config_filepath.exists() and not force:
        raise RuntimeError(
            f'Refusing to overwrite existing CONFIG, {config_filepath}, '
            'use --force if you really want to')

    logger.debug('Copying default config from %s to %s', src, config_filepath)

    op_queue.add(action=shutil.copy,
                 kwargs={
                     'src': src,
                     'dst': config_filepath,
                 },
                 description='Copying config to',
                 extra_description=f'{config_filepath}')



from __future__ import annotations

from getpass import getuser

from pydantic import Field, validator

from ._basemodel import BaseModel


class ImasBaseModel(BaseModel):
    """This model describes an IMAS data location."""
    user: str = Field(None, description='Username.')
    db: str = Field(description='IMAS db/machine name.')
    shot: int = Field(description='IMAS Shot number.')
    run: int = Field(description='IMAS Run number.')

    @validator('user', pre=True, always=True)
    def validate_user(cls, v):
        return v or getuser()



from __future__ import annotations

import logging
import os
from pathlib import Path
from typing import List, Sequence, Union

from importlib_resources import files

from ..schema import IDSVariableModel
from ..schema.variables import VariableConfigModel

logger = logging.getLogger(__name__)

VAR_ENV = 'DUQTOOLS_VARDEF'
USER_CONFIG_HOME = Path.home() / '.config'
LOCAL_DIR = Path('.').absolute()
DUQTOOLS_DIR = 'duqtools'
VAR_FILENAME = 'variables.yaml'


class VariableConfigLoader:

    def __init__(self):
        path = self.get_config_path()

        if not path.exists():
            raise OSError(f'{path} does not exist!')

        self.path = path

    def load(self):
        """Load the variables config."""
        logger.debug(f'Loading variables from: {self.path}')
        return VariableConfigModel.parse_file(self.path)

    def get_config_path(self) -> Path:
        """Try to get the config file with variable definitions.

        Search order:
        1. environment variable
        (2. local directory, not sure if this should be implemented)
        3. config home (first $XDG_CONFIG_HOME/duqtools then `$HOME/.config/duqtools`)
        4. fall back to variable definitions in package
        """
        for path in (
                self._get_path_from_environment_variable(),
                self._get_path_from_config_home(),
                self._get_path_local_directory(),
        ):
            if path:
                return path

        return self._get_path_fallback()

    def _get_path_from_environment_variable(self):
        env = os.environ.get(VAR_ENV)
        if env:
            test_path = Path(env)
            if not test_path.exists():
                raise OSError(
                    f'{test_path} defined by ${VAR_ENV} does not exist!')
            return test_path

    def _get_path_local_directory(self):
        return None  # Not implemented

    def _get_path_from_config_home(self):
        config_home = os.environ.get('XDG_CONFIG_HOME', USER_CONFIG_HOME)

        test_path = Path(config_home) / DUQTOOLS_DIR / VAR_FILENAME
        if test_path.exists():
            return test_path

    def _get_path_fallback(self):
        return files('duqtools.data') / VAR_FILENAME


def lookup_vars(
    variables: Sequence[Union[str,
                              IDSVariableModel]]) -> List[IDSVariableModel]:
    """Helper function to look up a bunch of variables.

    If str, look up the variable from the `var_lookup` Else check if the
    variable is an `IDSVariableModel`
    """
    var_models = []
    for var in variables:
        if isinstance(var, str):
            var = var_lookup[var]
        if not isinstance(var, IDSVariableModel):
            raise ValueError(f'Cannot lookup variable with type {type(var)}')
        var_models.append(var)
    return var_models


variable_config = VariableConfigLoader().load()
var_lookup = variable_config.to_variable_dict()



import sys
from pathlib import Path
from typing import Sequence

import streamlit as st
import xarray as xr

from duqtools.config import var_lookup
from duqtools.ids import (ImasHandle, rebase_on_grid, rebase_on_time,
                          standardize_grid)

try:
    default_workdir = sys.argv[1]
except IndexError:
    default_workdir = str(Path.cwd())

PREFIX = 'profiles_1d/*'
PREFIX0 = 'profiles_1d/0'


@st.experimental_memo
def get_ids_options():
    options = tuple(
        {v.ids
         for v in var_lookup.values() if (v.type == 'IDS-variable')})
    return options


@st.experimental_memo
def get_var_options(ids):
    options = [
        k for k, v in var_lookup.items()
        if (v.type == 'IDS-variable') and (v.ids == ids)
    ]
    return options


@st.experimental_memo
def get_variables(*, ids: str, x_key: str, y_keys: Sequence[str]):
    return {
        'time_var': 'time',
        'grid_var': x_key,
        'data_vars': y_keys,
    }


@st.experimental_memo
def _get_dataset(handles, *, ids, time_var, grid_var, data_vars):
    datasets = []

    variables = tuple((time_var, grid_var, *data_vars))

    for name, handle in handles.items():
        handle = ImasHandle(**handle)

        ds = handle.get_variables(variables=variables)

        ds = standardize_grid(
            ds,
            new_dim=grid_var,
            old_dim='x',
            new_dim_data=0,
            group=time_var,
        )
        datasets.append(ds)

    reference_grid = datasets[0][grid_var].data

    datasets = [
        rebase_on_grid(ds, coord_dim=grid_var, new_coords=reference_grid)
        for ds in datasets
    ]

    reference_time = datasets[0][time_var].data

    datasets = [
        rebase_on_time(ds, time_dim=time_var, new_coords=reference_time)
        for ds in datasets
    ]

    dataset = xr.concat(datasets, 'run')
    dataset['run'] = list(handles.keys())

    return dataset


def get_dataset(handles, *, ids, time_var, grid_var, data_vars):
    """Convert to hashable types before calling `_get_dataset`."""
    handles = {name: handle.dict() for name, handle in handles.items()}
    return _get_dataset(handles,
                        ids=ids,
                        time_var=time_var,
                        grid_var=grid_var,
                        data_vars=data_vars)



from __future__ import annotations

import logging
from typing import Union

import numpy as np

from .._logging_utils import duqlog_screen
from ..apply_model import apply_model
from ..schema import IDSOperation
from ._handle import ImasHandle
from ._mapping import IDSMapping

logger = logging.getLogger(__name__)


@apply_model.register  # type: ignore
def _apply_ids(model: IDSOperation, *,
               ids_mapping: Union[ImasHandle, IDSMapping], **kwargs) -> None:
    """_. Implementation for IDS operations.

    Parameters
    ----------
    model : IDSOperation
        model
    target_in : ImasHandle, IDSMapping
        ImasHandle/IDSHandle, data to apply operation to.

    Returns
    -------
    None
    """
    target_in = None
    if isinstance(ids_mapping, ImasHandle):
        target_in = ids_mapping
        ids_mapping = ids_mapping.get(model.variable.ids)

    if isinstance(model.variable, str):
        raise TypeError('`model.variable` must have a `path` attribute.')

    npfunc = getattr(np, model.operator)

    data_map = ids_mapping.findall(model.variable.path)

    if len(data_map) == 0:
        # TODO this should be 'raise Error' but that breaks our tests,
        # leave it as a message until we can fix it with a Jetto-IMAS container
        duqlog_screen.error(
            f'{model.variable.path} not found in IDS, cannot adjust value')

    for path, data in data_map.items():
        if model.scale_to_error:
            sigma_key = path + model._upper_suffix

            if model.value < 0:
                lower_key = path + model._lower_suffix
                if lower_key in ids_mapping:
                    sigma_key = lower_key

            if sigma_key not in ids_mapping:
                raise ValueError(f'scale_to_error={model.scale_to_error} '
                                 f'but `{sigma_key}` is empty.')

            sigma_bound = ids_mapping[sigma_key]
            sigma = abs(sigma_bound - data)

            value = sigma * model.value
        else:

            value = model.value

        logger.info('Apply %s', model)

        logger.debug('data range before: %s - %s', data.min(), data.max())
        npfunc(data, value, out=data)
        logger.debug('data range after: %s - %s', data.min(), data.max())

    if target_in:
        logger.info('Writing data entry: %s', target_in)
        ids_mapping.sync(target_in)



# https://setuptools.pypa.io/en/latest/pkg_resources.html#workingset-objects
def fix_dependencies():
    import __main__
    __main__.__requires__ = ['jetto_tools>=1.8.6']
    import pkg_resources  # noqa


fix_dependencies()

__author__ = 'Carbon Collective'
__email__ = 's.smeets@esciencecenter.nl'
__version__ = '1.1.1'

import logging  # noqa
import warnings  # noqa

logging.basicConfig(level=logging.INFO)

# https://github.com/duqtools/duqtools/issues/264
warnings.filterwarnings(
    'ignore',
    'Explicit custom root behavior not yet implemented for pydantic_yaml')



from typing import List, Union

from ._basemodel import BaseModel
from ._variable import IDSVariableModel, JettoVariableModel


class VariableConfigModel(BaseModel):
    __root__: List[Union[JettoVariableModel, IDSVariableModel]]

    def __iter__(self):
        yield from self.__root__

    def __getitem__(self, index: int):
        return self.__root__[index]

    def to_variable_dict(self) -> dict:
        """Return dict of variables."""
        return {variable.name: variable for variable in self}



from pathlib import Path
from typing import List

from pydantic import validator

from ..schema.runs import Run, Runs
from ..schema.workdir import WorkDirectoryModel


class WorkDirectory(WorkDirectoryModel):

    @property
    def cwd(self):
        cwd = Path.cwd()
        if not cwd.relative_to(self.root):
            raise OSError(
                f'Work directory must be a subdirectory of {self.root}')
        return cwd

    @property
    def subdir(self):
        """Get subdirectory relative to root."""
        return self.cwd.relative_to(self.root)

    @validator('root')
    def resolve_root(cls, v):
        return v.resolve()

    @property
    def runs_yaml(self):
        """Location of runs.yaml."""
        return self.cwd / 'runs.yaml'

    @property
    def runs_yaml_old(self):
        """Location of runs.yaml.old."""
        return self.cwd / 'runs.yaml.old'

    @property
    def runs(self) -> List[Run]:
        """Get a list of the runs currently created from this config."""
        runs_yaml = self.runs_yaml

        if not runs_yaml.exists():
            raise OSError(f'Cannot find {runs_yaml}.')

        return Runs.parse_file(runs_yaml)
